/* autoconfig.h.  Generated from autoconfig.h.in by configure.  */
/* autoconfig.h.in.  Generated from configure.in by autoheader.  */

/* Define if building universal (internal helper macro) */
/* #undef AC_APPLE_UNIVERSAL_BUILD */

/* Define to path for platform specific include file */
/* #undef ADDON_FILE */

/* Define to 1 if compiler threads static member derived from same base thru
   multiple paths as ambigious */
/* #undef AMBIGIOUS_NEW_BUG */

/* Define for short name of hardware platform */
/* #undef HARDWARE */

/* Define to 1 if __attribute((aligned)) works */
#define HAS_ALIGNED 1

/* Define to 1 if the GNU __attribute__ extension is available */
#define HAS_ATTRIBUTES 1

/* Define to 1 if the variable CLK_TCK is available */
/* #undef HAS_CLK_TCK */

/* Define to 1 if const_cast<> works */
#define HAS_CONST_CAST 1

/* Define to 1 if the variable HZ is available */
#define HAS_HZ 1

/* Define to 1 if the C99 type int16_t is available */
#define HAS_INT16_T 1

/* Define to 1 if the C99 type int32_t is available */
#define HAS_INT32_T 1

/* Define to 1 if the C99 type int64_t is available */
#define HAS_INT64_T 1

/* Define to 1 if the C99 type int8_t is available */
#define HAS_INT8_T 1

/* Define to 1 if integer constants are valid template arguments */
#define HAS_INT_TEMPLATE_ARGS 1

/* Define to 1 if templated subclasses are correctly in the namespace of the
   parent class */
#define HAS_LOCAL_TEMPLATES 1

/* Define to 1 if __attribute((may_alias)) works */
#define HAS_MAY_ALIAS 1

/* Define to 1 if initializers of static const integral members work */
#define HAS_MEMBER_INIT 1

/* Define to 1 if subclasses have access to private members */
#define HAS_PRIVATE_ACCESS 1

/* Define to 1 if ptrdiff_t type is available */
#define HAS_PTRDIFF_T 1

/* Define to 1 if reinterpret_cast<> works */
#define HAS_REINTERPRET_CAST 1

/* Define to 1 if _SC_CLK_TCK is available for sysconf */
#define HAS_SC_CLK_TCK 1

/* Define to 1 if the STDERR_FILENO define is available */
#define HAS_STDERR_FILENO 1

/* Define to 1 if the STDIN_FILENO define is available */
#define HAS_STDIN_FILENO 1

/* Define to 1 if the STDOUT_FILENO define is available */
#define HAS_STDOUT_FILENO 1

/* Define to 1 if struct timeval is available */
#define HAS_STRUCT_TIMEVAL 1

/* Define to 1 if struct tms is available */
#define HAS_STRUCT_TMS 1

/* Define to 1 if casting in template arguments work */
#define HAS_TEMPLATE_CASTS 1

/* Define to 1 if the C99 type uint16_t is available */
#define HAS_UINT16_T 1

/* Define to 1 if the C99 type uint32_t is available */
#define HAS_UINT32_T 1

/* Define to 1 if the C99 type uint64_t is available */
#define HAS_UINT64_T 1

/* Define to 1 if the C99 type uint8_t is available */
#define HAS_UINT8_T 1

/* Define to 1 if __attribute((visibility("default"))) works */
#define HAS_VISIBILITY_DEFAULT 1

/* Define to 1 if __attribute((visibility("hidden"))) works */
#define HAS_VISIBILITY_HIDDEN 1

/* Define to 1 if __attribute((visibility("internal"))) works */
#define HAS_VISIBILITY_INTERNAL 1

/* Define to 1 if the __null keyword is available */
#define HAS__NULL_TYPE 1

/* Define if additional settings are required for this architecture */
/* #undef HAVE_ADDONS */

/* Define to 1 if the always_inline attribute is available */
/* #undef HAVE_ALWAYS_INLINE */

/* Define to 1 if you have the <assert.h> header file. */
#define HAVE_ASSERT_H 1

/* Define to 1 if __sync arithmetic functions are available */
#define HAVE_ATOMIC_ADDSUB 1

/* Define to 1 if __sync_val_compare_and_swap is available */
#define HAVE_ATOMIC_COMPARE_AND_SWAP 1

/* Define to 1 if you have the <bstrings.h> header file. */
/* #undef HAVE_BSTRINGS_H */

/* Define to 1 if you have the <bstring.h> header file. */
/* #undef HAVE_BSTRING_H */

/* Define to 1 if __builtin_expect is available */
#define HAVE_BUILTIN_EXPECT 1

/* Define to 1 if __builtin_memcpy is available */
#define HAVE_BUILTIN_MEMCPY 1

/* Define to 1 if __builtin_memmove is available */
#define HAVE_BUILTIN_MEMMOVE 1

/* Define to 1 if __builtin_memset is available */
#define HAVE_BUILTIN_MEMSET 1

/* Define to 1 if you have the `clock' function. */
#define HAVE_CLOCK 1

/* Define to 1 if you have the `close' function. */
#define HAVE_CLOSE 1

/* Define to 1 if you have the <ctype.h> header file. */
#define HAVE_CTYPE_H 1

/* Define to 1 if you don't have `vprintf' but do have `_doprnt.' */
/* #undef HAVE_DOPRNT */

/* Define to 1 if you have the `dup' function. */
#define HAVE_DUP 1

/* Define to 1 if you have the `dup2' function. */
#define HAVE_DUP2 1

/* Define to 1 if you have the <errno.h> header file. */
#define HAVE_ERRNO_H 1

/* Define to 1 if you have the `execve' function. */
#define HAVE_EXECVE 1

/* Define to 1 if you have the <fcntl.h> header file. */
#define HAVE_FCNTL_H 1

/* Define to 1 to use the autoconfig feature selection */
/* #undef HAVE_FEATURE_SELECTION */

/* Define to 1 if float SIMD data types are available */
/* #undef HAVE_FLOAT_SIMD */

/* Define to 1 if you have the `fopen64' function. */
#define HAVE_FOPEN64 1

/* Define to 1 if you have the `fork' function. */
#define HAVE_FORK 1

/* Define to 1 if you have the `free' function. */
#define HAVE_FREE 1

/* Define to 1 if you have the `fstat' function. */
#define HAVE_FSTAT 1

/* Define to 1 if you have the `gettimeofday' function. */
#define HAVE_GETTIMEOFDAY 1

/* Define to 1 if you have the `htonl' function. */
#define HAVE_HTONL 1

/* Define to 1 if you have the `htons' function. */
#define HAVE_HTONS 1

/* Define to 1 if you have the <inttypes.h> header file. */
#define HAVE_INTTYPES_H 1

/* Define to 1 if you have the `isspace' function. */
#define HAVE_ISSPACE 1

/* Define to 1 if you have the `kill' function. */
#define HAVE_KILL 1

/* Define to 1 if llseek is available */
/* #undef HAVE_LLSEEK */

/* Define to 1 if you have the `longjmp' function. */
#define HAVE_LONGJMP 1

/* Define to 1 if the system has the type `long long'. */
#define HAVE_LONG_LONG 1

/* Define to 1 if you have the `lseek' function. */
#define HAVE_LSEEK 1

/* Define to 1 if lseek64 is available */
#define HAVE_LSEEK64 1

/* Define to 1 if you have the `malloc' function. */
#define HAVE_MALLOC 1

/* Define to 1 if you have the <math.h> header file. */
#define HAVE_MATH_H 1

/* Define to 1 if you have the `memchr' function. */
#define HAVE_MEMCHR 1

/* Define to 1 if you have the `memmove' function. */
#define HAVE_MEMMOVE 1

/* Define to 1 if you have the <memory.h> header file. */
#define HAVE_MEMORY_H 1

/* Define to 1 if you have the `memset' function. */
#define HAVE_MEMSET 1

/* Define to 1 if you have the <netinet/in.h> header file. */
#define HAVE_NETINET_IN_H 1

/* Define to 1 if the noreturn attribute is available */
#define HAVE_NORETURN 1

/* Define to 1 if you have the `ntohl' function. */
#define HAVE_NTOHL 1

/* Define to 1 if you have the `ntohs' function. */
#define HAVE_NTOHS 1

/* Define to 1 if you have the `open' function. */
#define HAVE_OPEN 1

/* Define to 1 if you have the `pipe' function. */
#define HAVE_PIPE 1

/* Define if you have POSIX threads libraries and header files. */
/* #undef HAVE_PTHREAD */

/* Define to 1 if you have the `pthread_create' function. */
/* #undef HAVE_PTHREAD_CREATE */

/* Define to 1 if PTHREAD_CREATE_DETACHED is available */
/* #undef HAVE_PTHREAD_CREATE_DETACHED */

/* Define to 1 if you have the `pthread_detach' function. */
/* #undef HAVE_PTHREAD_DETACH */

/* Define to 1 if you have the `pthread_equal' function. */
/* #undef HAVE_PTHREAD_EQUAL */

/* Define to 1 if you have the <pthread.h> header file. */
/* #undef HAVE_PTHREAD_H */

/* Define to 1 if you have the `pthread_mutexattr_init' function. */
/* #undef HAVE_PTHREAD_MUTEXATTR_INIT */

/* Define to 1 if you have the `pthread_mutex_init' function. */
/* #undef HAVE_PTHREAD_MUTEX_INIT */

/* Define to 1 if you have the `pthread_self' function. */
/* #undef HAVE_PTHREAD_SELF */

/* Define to 1 if you have the `read' function. */
#define HAVE_READ 1

/* Define to 1 if you have the `rename' function. */
#define HAVE_RENAME 1

/* Define to 1 if the __restrict__ pointer attribute is available */
#define HAVE_RESTRICTED_PTRS 1

/* Define to 1 if you have the <semaphore.h> header file. */
/* #undef HAVE_SEMAPHORE_H */

/* Define to 1 if you have the `sem_init' function. */
/* #undef HAVE_SEM_INIT */

/* Define to 1 if you have the <setjmp.h> header file. */
#define HAVE_SETJMP_H 1

/* Define to 1 if you have the `sigaction' function. */
#define HAVE_SIGACTION 1

/* Define to 1 if you have the `sigemptyset' function. */
#define HAVE_SIGEMPTYSET 1

/* Define to 1 if you have the `signal' function. */
#define HAVE_SIGNAL 1

/* Define to 1 if you have the <signal.h> header file. */
#define HAVE_SIGNAL_H 1

/* Define to 1 if you have the `sleep' function. */
#define HAVE_SLEEP 1

/* Define to 1 if you have the `snprintf' function. */
#define HAVE_SNPRINTF 1

/* Define to 1 if you have the <stdarg.h> header file. */
#define HAVE_STDARG_H 1

/* Define to 1 if you have the <stddef.h> header file. */
#define HAVE_STDDEF_H 1

/* Define to 1 if you have the <stdint.h> header file. */
#define HAVE_STDINT_H 1

/* Define to 1 if you have the <stdio.h> header file. */
#define HAVE_STDIO_H 1

/* Define to 1 if you have the <stdlib.h> header file. */
#define HAVE_STDLIB_H 1

/* Define to 1 if you have the `strchr' function. */
#define HAVE_STRCHR 1

/* Define to 1 if you have the `strerror' function. */
#define HAVE_STRERROR 1

/* Define to 1 if you have the <strings.h> header file. */
#define HAVE_STRINGS_H 1

/* Define to 1 if you have the <string.h> header file. */
#define HAVE_STRING_H 1

/* Define to 1 if you have the `strrchr' function. */
#define HAVE_STRRCHR 1

/* Define to 1 if you have the `strtod' function. */
#define HAVE_STRTOD 1

/* Define to 1 if you have the `strtol' function. */
#define HAVE_STRTOL 1

/* Define to 1 if you have the `sysconf' function. */
#define HAVE_SYSCONF 1

/* Define to 1 if you have the `system' function. */
#define HAVE_SYSTEM 1

/* Define to 1 if you have the <sys/param.h> header file. */
#define HAVE_SYS_PARAM_H 1

/* Define to 1 if you have the <sys/stat.h> header file. */
#define HAVE_SYS_STAT_H 1

/* Define to 1 if you have the <sys/times.h> header file. */
#define HAVE_SYS_TIMES_H 1

/* Define to 1 if you have the <sys/time.h> header file. */
#define HAVE_SYS_TIME_H 1

/* Define to 1 if you have the <sys/types.h> header file. */
#define HAVE_SYS_TYPES_H 1

/* Define to 1 if you have the `time' function. */
#define HAVE_TIME 1

/* Define to 1 if you have the `times' function. */
#define HAVE_TIMES 1

/* Define to 1 if you have the <time.h> header file. */
#define HAVE_TIME_H 1

/* Define to 1 if you have the <unistd.h> header file. */
#define HAVE_UNISTD_H 1

/* Define to 1 if the system has the type `unsigned long long'. */
#define HAVE_UNSIGNED_LONG_LONG 1

/* Define to 1 if you have the `vprintf' function. */
#define HAVE_VPRINTF 1

/* Define to 1 if you have the `vsnprintf' function. */
#define HAVE_VSNPRINTF 1

/* Define to 1 if you have the `write' function. */
#define HAVE_WRITE 1

/* Define to 1 if the system has the type `__int64'. */
/* #undef HAVE___INT64 */

/* Define to 1 for big endian */
/* #undef JPG_BIG_ENDIAN */

/* Define to 1 for little endian */
#define JPG_LIL_ENDIAN 1

/* Define to the exponent of a power of two giving the worst-case alignment of
   components */
#define NATURAL_ALIGNMENT 4

/* Define to the address where bug reports for this package should be sent. */
#define PACKAGE_BUGREPORT ""

/* Define to the full name of this package. */
#define PACKAGE_NAME "jpeg"

/* Define to the full name and version of this package. */
#define PACKAGE_STRING "jpeg thomas.richter@iis.fraunhofer.de"

/* Define to the one symbol short name of this package. */
#define PACKAGE_TARNAME "jpeg"

/* Define to the home page for this package. */
#define PACKAGE_URL ""

/* Define to the version of this package. */
#define PACKAGE_VERSION "thomas.richter@iis.fraunhofer.de"

/* Define to the necessary symbol if this constant uses a non-standard name on
   your system. */
/* #undef PTHREAD_CREATE_JOINABLE */

/* Define as the return type of signal handlers (`int' or `void'). */
#define RETSIGTYPE void

/* Define to 1 if required argument to signal handlers is "..." */
/* #undef SIG_ARG_TYPE_DOTS */

/* Define to 1 if required argument to signal handlers is int */
#define SIG_ARG_TYPE_INT 1

/* The size of `char', as computed by sizeof. */
#define SIZEOF_CHAR 1

/* The size of `int', as computed by sizeof. */
#define SIZEOF_INT 4

/* The size of `long', as computed by sizeof. */
#define SIZEOF_LONG 8

/* The size of `long long', as computed by sizeof. */
#define SIZEOF_LONG_LONG 8

/* The size of `short', as computed by sizeof. */
#define SIZEOF_SHORT 2

/* The size of `void *', as computed by sizeof. */
#define SIZEOF_VOID_P 8

/* The size of `__int64', as computed by sizeof. */
#define SIZEOF___INT64 0

/* Define to 1 if the `S_IS*' macros in <sys/stat.h> do not work properly. */
/* #undef STAT_MACROS_BROKEN */

/* Define to 1 if you have the ANSI C header files. */
#define STDC_HEADERS 1

/* Define to 1 if you can safely include both <sys/time.h> and <time.h>. */
#define TIME_WITH_SYS_TIME 1

/* Define if you want to use multithreading. */
/* #undef USE_MULTITHREADING */

/* Define WORDS_BIGENDIAN to 1 if your processor stores words with the most
   significant byte first (like Motorola and SPARC, unlike Intel). */
#if defined AC_APPLE_UNIVERSAL_BUILD
# if defined __BIG_ENDIAN__
#  define WORDS_BIGENDIAN 1
# endif
#else
# ifndef WORDS_BIGENDIAN
/* #  undef WORDS_BIGENDIAN */
# endif
#endif

/* Define to empty if `const' does not conform to ANSI C. */
/* #undef const */

/* Define to `__inline__' or `__inline' if that's what the C compiler
   calls it, or to nothing if 'inline' is not supported under any name.  */
#ifndef __cplusplus
/* #undef inline */
#endif

/* Define to `unsigned int' if <sys/types.h> does not define. */
/* #undef size_t */
